import requests
from datetime import datetime
# from xml.etree import ElementTree

import xmltodict
from dateutil.relativedelta import relativedelta


class PortfolioManager:
    def __init__(self, username, password):
        self._headers = {'Content-Type': 'application/xml'}
        self._base_url = "https://portfoliomanager.energystar.gov/ws"
        self.username = username
        self.password = password

    def get_portfolio_manager_url(self, endpoint):
        """Points the request to the Portfolio Manager API. 
        Will not be useful if used on its own."""
        return self.handle_pm_html(
            requests.get("{}{}".format(self._base_url, endpoint),
                         auth=(self.username, self.password),
                         headers=self._headers))

    @staticmethod
    def handle_pm_html(response):
        """Consistent way to handle HTML in case a 
        call does not get an XML response."""
        if response.raise_for_status():
            return response.raise_for_status()
        else:
            return xmltodict.parse(response.text)

    # Account Information

    @property
    def account_info(self):
        """Get basic account information (name, email, address, phone, etc)"""
        return self.get_portfolio_manager_url('/account')

    @property
    def account_id(self):
        """Returns the Portfolio Manager ID associated
        with the active account"""
        account_info = self.account_info
        account_id = account_info['account']['id']
        return account_id

    @property
    def account_shares(self):
        """Returns a list of accounts that have requested 
        sharing their account information with Enertiv. """
        return self.get_portfolio_manager_url('/connect/account/pending/list')

    @property
    def property_shares(self):
        """Returns a list of property share requests that are pending. 
        These property share requests belong to customers that you 
        are already connected to."""
        return self.get_portfolio_manager_url('/share/property/pending/list')

    @property
    def meter_shares(self):
        """Returns a list of property share requests that are pending. 
        These property share requests belong to customers that you 
        are already connected to."""
        return self.get_portfolio_manager_url('/share/meter/pending/list')

    @property
    def portfolio_manager_notifications(self):
        """Returns a general list of notifications from Portfolio Manager. 
        Useful to check after each login to make sure no shares or meters 
        have been disconnected, or that energy data  has been correctly 
        put into Portfolio Manager."""
        return self.get_portfolio_manager_url("/notification/list")

    @property
    def customer_list(self):
        """Returns a full list of accounts that have connected their Portfolio 
        Manager account to Enertiv's Portfolio Manager account. An account 
        level share enables a user to share properties, but does not guarantee 
        that any given property will be shared with Enertiv."""
        return self.get_portfolio_manager_url('/customer/list')

    def customer_property_list(self, customer_account_id):
        """Returns a list of Property objects. The list contains all buildings 
        shared with the active Portfolio Manager account for a given 
        customer account ID."""
        properties = self.get_portfolio_manager_url(
            "/account/{}/property/list".format(str(customer_account_id)))

        # Stop method and return HTTP response if call does not go as expected.
        if properties['response']['@status'] != 'Ok':
            return properties

        # Create a single list of Property objects to return.
        property_list = []
        try:
            for customer_property in properties['response']['links'].values():
                property_as_object = Property(
                    customer_property['@id'], self.username, self.password)
                property_list.append(property_as_object)
        except AttributeError:
            # AttributeError occurs when a customer has no buildings.
            pass
        return property_list

    @property
    def full_building_list(self):
        """Returns a list of all buildings
        currently connected to your account."""
        full_customer_list = []
        building_list = []

        for customer in self.customer_list['response']['links']['link']:
            try:
                # store the customer name and ID, ignore the rest
                full_customer_list.append((customer['@hint'], customer['@id']))
            except:
                pass
        try:
            for shared_account in full_customer_list:
                # print the name and ID for each account
                # print('{}: {}'.format(shared_account[0], shared_account[1]))
                # for each account, store shared buildings and IDs
                property_list = self.customer_property_list(shared_account[1])
                [building_list.append(x) for x in property_list]
        except AttributeError:
            # AttributeError occurs when a customer has no buildings.
            pass

        return building_list

    # ==========================================================#
    #  Functions below are for pushing data into Portfolio Manager
    # (NOTE: Not fully functional yet -- Andrew)
    #
    # def _form_xml(self, action):
    #     response = ElementTree.Element('sharingResponse')
    #     act = ElementTree.SubElement(response, 'action')
    #     act.text = action
    #     return ElementTree.tostring(response)
    #
    # def handle_share(self, action, customer_id):
    #     """Handles an incoming share in Portfolio Manager.
    #     Accepted inputs are "Accept" and "Reject."""
    #     handling_xml = self._form_xml(action)
    #     handle_share = requests.post("{}/connect/account/{}"
    #                                  .format(self._base_url, customer_id),
    #                                  auth=(self.username, self.password),
    #                                  headers=self._headers, data=handling_xml)
    #     return xmltodict.parse(handle_share.text)


class Property(PortfolioManager):
    def __init__(self, property_id, username, password):
        PortfolioManager.__init__(self, username, password)
        self.property_id = property_id

    @property
    def property_info(self):
        """Provide a property ID (retrievable from the property_list method) 
        and get basic information about a particular property."""
        return self.get_portfolio_manager_url(
            '/property/{}'.format(str(self.property_id)))

    @property
    def property_use_list(self):
        """Returns a list of all property uses within a property"""
        property_uses = self.get_portfolio_manager_url(
            "/property/{}/propertyUse/list".format(str(self.property_id)))

        # Stop method and return HTTP response if call does not go as expected.
        if property_uses['response']['@status'] != 'Ok':
            return property_uses

        # Create a list of PropertyUse objects to return.
        property_use_list = []
        for property_use in property_uses['response']['links']['link']:
            property_use_as_object = PropertyUse(
                property_use['@id'], self.property_id,
                self.username, self.password)
            property_use_list.append(property_use_as_object)
        return property_use_list

    @property
    def meter_list(self):
        """Returns a list of meters associated with the property"""
        meters = self.get_portfolio_manager_url(
            '/property/{}/meter/list'.format(str(self.property_id)))

        # Stop method and return HTTP response if call does not go as expected.
        if meters['response']['@status'] != 'Ok':
            return meters

        # Create a list of Meter objects to return.
        meter_list = []
        for property_meter in meters['response']['links']['link']:
            meter_as_object = Meter(
                property_meter['@id'], self.property_id,
                self.username, self.password)
            meter_list.append(meter_as_object)
        return meter_list

    # Metrics (e.g. ENERGY STAR Score)

    def get_ENERGY_STAR_score(
            self, year=datetime.now().year, month=datetime.now().month):
        """Get an ENERGY STAR score for a particular year/month. 
        Default is to check the most recent month. """
        ES_score = requests.get('{}/property/{}/metrics?year={}&month={}&'
            'measurementSystem=EPA'.format(self._base_url,
                                           str(self.property_id), str(year),
                                           str(month)),
                                auth=(self.username, self.password),
                                headers={**self._headers,
                                         'PM-Metrics': 'score'})
        return self.handle_pm_html(ES_score)

    def get_metrics(self, year=datetime.now().year, month=datetime.now().month,
                    metrics='score, medianScore, siteTotal, sourceTotal, '
                            'siteIntensity, sourceTotalWN, '
                            'indirectGHGEmissions, '
                            'coolingDegreeDaysCDD, heatingDegreeDaysHDD, '
                            'energyStarCertificationEligibility'):
        """Get metrics from Portfolio Manager for a particular year/month. 
        The default is to check the most recent month and provide a sample of 
        popular metrics. A full list of available metrics is available at 
        http://downloads.energystar.gov/bi/portfolio-manager/Custom_Reporting_Metric_Inventory_en_US.xlsx"""

        many_metrics = requests.get(
            "{}/property/{}/metrics?year={}&month={}&measurementSystem=EPA"
            .format(self._base_url,
                    str(self.property_id),
                    str(year), str(month)),
            auth=(self.username, self.password),
            headers={**self._headers,
                     'PM-Metrics': metrics})
        return self.handle_pm_html(many_metrics)

    def get_ENERGY_STAR_scores(self, start_date='1999-01',
                               end_date=datetime.today().strftime("%Y-%m")):
        """A list of all available ENERGY STAR scores over a given time frame.
        Default time frame is between today and January 1, 1999."""

        ENERGY_STAR_scores = []

        start_date = datetime.strptime(start_date, '%Y-%m')
        end_date = datetime.strptime(end_date, '%Y-%m')
        months_back = 0

        while end_date + relativedelta(months=-months_back) >= start_date:

            year = str((end_date + relativedelta(months=-months_back)).year)
            month = str((end_date + relativedelta(months=-months_back)).month)

            score = self.get_ENERGY_STAR_score(str(year), str(month))[
                'propertyMetrics']['metric']['value']
            if isinstance(score, dict):
                pass
            else:
                ENERGY_STAR_scores.append((month, year, score))

            months_back += 1
        return ENERGY_STAR_scores


class PropertyUse(Property):
    def __init__(self, property_use_id, property_id, username, password):
        Property.__init__(self, property_id, username, password)
        self.property_use_id = property_use_id

    @property
    def get_property_use_info(self):
        """Basic information about a single property use, including a
        property use detail IDs"""
        return self.get_portfolio_manager_url(
            "/propertyUse/{}".format(str(self.property_use_id)))

    def get_property_use_details(self, use_detail_id):
        """Returns information about a single property use detail."""
        return self.get_portfolio_manager_url(
            "/useDetails/{}".format(str(use_detail_id)))

    @property
    def get_hierarchy(self):
        """Get the account ID and property ID
        associated with the property use."""
        return self.get_portfolio_manager_url(
            '/idHierarchy/propertyUse/{}'.format(self.property_use_id))

    @property
    def meter_list(self):
        """Returns a list of meters associated with the property use.
        Many buildings will not have any meters
        assigned to specific property uses."""
        meters = self.get_portfolio_manager_url(
            '/association/propertyUse/{}/meter'
                .format(str(self.property_use_id)))

        # Stop method and return HTTP response if call does not go as expected.
        if meters['response']['@status'] != 'Ok':
            return meters

        # Create a list of Meter objects to return.
        meter_list = []
        for property_meter in meters['response']['links']['link']:
            meter_as_object = Meter(
                property_meter['@id'], self.property_id,
                self.username, self.password)
            meter_list.append(meter_as_object)
        return meter_list


class Meter(Property):
    def __init__(self, meter_id, property_id, username, password):
        Property.__init__(self, property_id, username, password)
        self.meter_id = meter_id

    @property
    def meter_info(self):
        """Provide a meter ID and get basic information about a meter"""
        return self.get_portfolio_manager_url(
            "/meter/{}".format(str(self.meter_id)))

    def meter_consumption_data_recent(self, start_date='1999-01-01',
                                      end_date=datetime.today().strftime(
                                          '%Y-%m-%d')):
        """Provide a meter ID and get consumption data over a given time frame.
        The default time frame is the most recent 20 meter entries
        (typically, there is one entry per month)."""
        return self.get_portfolio_manager_url(
            "/meter/{}/consumptionData?startDate={}&endDate={}".format(
                str(self.meter_id), start_date, end_date))

    def meter_consumption_data_full(self, start_date='1999-01-01',
                                    end_date=datetime.today().strftime(
                                        '%Y-%m-%d')):
        """Provide a meter ID and get all available consumption data
        over a time frame, not limited to 20 entries. The default time frame
        is between today and January 1, 1999."""
        meter_data = []
        page_number = 1

        for page in range(100):
            long_url = "{}/meter/{}/consumptionData?startDate={}&endDate={}&page={}"\
                .format(start_date, end_date, self._base_url,
                        str(self.meter_id), page_number)
            meter_call = requests.get(long_url,
                                      auth=(self.username, self.password),
                                      headers=self._headers)
            response = xmltodict.parse(meter_call.text)

            try:
                meter_data.append(response['meterData']['meterConsumption'])
            except Exception:
                pass
            page_number += 1

        return meter_data
